from .Database import DatabaseManager
import discord


database_keys = ['guild', 'member', 'currency', 'bank']


class EconomyAccount:
    def __init__(self, guild: int, member: int, database: DatabaseManager, table):
        self.guild = guild
        self.member = member
        self.database = database
        self.table = table

    def __str__(self):
        return f"<Account MEMBER={self.member}, GUILD={self.guild}>"

    def __repr__(self):
        return f"<Account GUILD={self.guild}, MEMBER={self.member}, CURRENCY={self.currency}, BANK={self.bank}>"

    def __lt__(self, other):
        return self.net < other.net

    @property
    def __checks(self):
        return EconomyManager.generate_checks(self.guild, self.member)

    @property
    def currency(self):
        return self.database.select(['currency'], self.table, self.__checks)[0]

    @property
    def bank(self):
        return self.database.select(['bank'], self.table, self.__checks)[0]

    @property
    def net(self):
        return self.bank + self.currency

    def change_currency(self, amount: int):
        self.database.update(['currency'], [self.currency + amount], self.table, self.__checks)

    def change_bank(self, amount: int):
        self.database.update(['bank'], [self.bank + amount], self.table, self.__checks)


class EconomyManager:
    def __init__(self, database: DatabaseManager, table, bot):
        self.database = database
        self.table = table
        self.bot = bot
        self.database.createtable(self.table, [{'name': key, 'type': 'INTEGER'} for key in database_keys], True)

    @staticmethod
    def generate_checks(guild: int, member: int):
        return [{'guild': guild}, {'member': member}]

    async def create_account(self, member: discord.Member):
        self.database.insertifnotexists(database_keys, [member.guild.id, member.id, 0, 0],
                                        self.table, self.generate_checks(member.guild.id, member.id))

    async def get_account(self, member: discord.Member):
        member_data = self.select([], table_name, generate_checks(member.guild.id, member.id), True)

        if member_data:
            return EconomyAccount(member.guild.id, member.id, self.database, self.table)

        return None

    async def get_leaderboard(self, guild):
        guild_info = self.database.select([], self.table, [{'guild': guild.id}], True)
        members = [EconomyAccount(*member_info[:2], self.database, self.table) for member_info in guild_info]

        members.sort()
        return members



import aiohttp
import discord
import youtube_dl
from .Base import EventManager
from typing import Optional
import asyncio
from enum import Enum

# should be .Base, koyashie use Base for testing

# just some options etc.

ytdl_opts = {
    'format': 'bestaudio/best',
    'restrictfilenames': True,
    'noplaylist': False,
    'nocheckcertificate': True,
    'ignoreerrors': False,
    'logtostderr': False,
    'quiet': True,
    'no_warnings': True,
    'default_search': 'auto'
}

ffmpeg_options = {
    'before_options': '-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5', 'options': '-vn'
}

ytdl = youtube_dl.YoutubeDL(ytdl_opts)


# errors/exceptions

class NotPlaying(Exception):
    """Raises error when client is not playing"""


class AlreadyPlaying(Exception):
    """Raises error when player is already playing"""


class NotConnected(Exception):
    """Raises error when client is not connected to a voice channel"""


class NotPaused(Exception):
    """Raises error when player is not paused"""


class QueueEmpty(Exception):
    """Raises error when queue is empty"""


class AlreadyConnected(Exception):
    """Raises error when client is already connected to voice"""


class AlreadyPaused(Exception):
    """Raises error when player is already paused."""


class QueueError(Exception):
    """Raises error when something is wrong with the queue"""


class SkipError(Exception):
    """Raises error when there is no song to skip to"""


class UserNotConnected(Exception):
    """Raises error when user is not connected to channel"""


class InvalidSkipIndex(Exception):
    """Raises error when the skip index is < 0"""


class Loops(Enum):
    NO_LOOP = 0
    LOOP = 1
    QUEUE_LOOP = 2


class Player(discord.PCMVolumeTransformer):
    def __init__(self, source, *, data, volume=0.1):
        super().__init__(source, volume)

        self.data = data
        self.title = data.get('title')
        self.url = data.get('url')
        self.duration = data.get('duration') if data.get('duration') != 0 else "LIVE"

    def __str__(self):
        return self.title

    @classmethod
    async def make_player(cls, query: str):
        data = await MusicManager.fetch_data(query)
        if data is None:
            await self.call_event('on_music_error', ctx, FetchFailed("Failed to fetch query data."))
            return []

        if 'entries' in data:
            return [cls(discord.FFmpegPCMAudio(player['url'],
                                               **ffmpeg_options), data=player) for player in data['entries']]

        filename = data['url']
        return cls(discord.FFmpegPCMAudio(filename, **ffmpeg_options), data=data)


class QueueManager:
    def __init__(self, volume, queue):
        self.queue = queue
        self.volume = volume
        self.history = []
        self.loop = Loops.NO_LOOP
        self.now_playing = None

    def add(self, player):
        self.queue.append(player)

    def clear(self):
        self.queue.clear()

    def remove(self, index):
        return self.queue.pop(index)


class MusicManager(EventManager):
    def __init__(self, bot):
        super().__init__()
        self.bot = bot
        self.queue = {}

    async def __check_connection(self, ctx, check_playing: bool = False, check_queue: bool = False) -> Optional[bool]:
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        if check_playing and not ctx.voice_client.is_playing():
            await self.call_event('on_music_error', ctx, NotPlaying("Client is not playing anything currently"))
            return

        if check_queue and ctx.guild.id not in self.queue:
            await self.call_event('on_music_error', ctx, QueueEmpty("Queue is empty"))
            return

        return True

    def __check_queue(self, ctx):
        try:
            if self.queue[ctx.guild.id].loop == Loops.LOOP:
                song = self.queue[ctx.guild.id].now_playing
                player = Player(discord.FFmpegPCMAudio(song.url, **ffmpeg_options), data=song.data)

            elif self.queue[ctx.guild.id].loop == Loops.QUEUE_LOOP:
                player = self.queue[ctx.guild.id].remove(0)
                self.queue[ctx.guild.id].add(player)

                player = Player(discord.FFmpegPCMAudio(player.url, **ffmpeg_options), data=player.data)

            else:
                player = self.queue[ctx.guild.id].remove(0)

            self.queue[ctx.guild.id].now_playing = player

            if player is not None and ctx.voice_client:
                player.volume = self.queue[ctx.guild.id].volume
                ctx.voice_client.play(player, after=lambda x: self.__check_queue(ctx))

                if self.queue[ctx.guild.id].loop == Loops.NO_LOOP:
                    self.queue[ctx.guild.id].history.append(player)
                    self.bot.loop.create_task(
                        self.call_event('on_play', ctx, player)
                    )
        except (IndexError, KeyError):
            return

    @staticmethod
    async def fetch_data(query: str):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, lambda: ytdl.extract_info(query, download=False))

    @staticmethod
    async def create_player(query):
        return await Player.make_player(query)

    async def queue_add(self, player, ctx):
        """Adds specified player object to queue"""

        if ctx.guild.id in self.queue:
            for song in player:
                self.queue[ctx.guild.id].add(song)
        else:
            self.queue[ctx.guild.id] = QueueManager(0.1, player)

    async def queue_remove(self, ctx, index):
        """Removed specified player object from queue"""

        if not await self.__check_connection(ctx, check_queue=True):
            return

        try:
            await self.queue[ctx.guild.id].remove(index)
        except IndexError:
            await self.call_event('on_music_error', ctx, QueueError("Failure when removing player from queue"))

    async def lyrics(self, ctx, query=None):
        query = self.now_playing(ctx) if query is None else query
        url = f"https://some-random-api.ml/lyrics?title={query}"

        async with aiohttp.ClientSession() as session:
            request = await session.get(url)
            request_json = await request.json()

            return await request_json.get('lyrics', None)

    async def play(self, ctx, player=None):
        if not await self.__check_connection(ctx):
            return

        if player is not None:
            ctx.voice_client.play(player)
            return True

        if not ctx.voice_client.is_playing():
            self.__check_queue(ctx)
            return True

    async def pause(self, ctx):
        if not await self.__check_connection(ctx):
            return

        if ctx.voice_client.is_paused():
            await self.call_event('on_music_error', ctx, AlreadyPaused("Player is already paused."))
            return

        ctx.voice_client.pause()
        return True

    async def resume(self, ctx):
        if not await self.__check_connection(ctx):
            return

        if not ctx.voice_client.is_paused():
            await self.call_event('on_music_error', ctx, NotPaused("Player is not paused"))
            return

        ctx.voice_client.resume()
        return True

    async def skip(self, ctx, index=None):
        if not await self.__check_connection(ctx, True, check_queue=True):
            return

        # Created duplicate to make sure InvalidSkipIndex isn't raised when the user does pass an index and the queue
        # is empty.
        skip_index = 0 if index is None else index - 1
        if not -1 < skip_index < len(self.queue[ctx.guild.id].queue):
            if index:
                await self.call_event('on_music_error', ctx, InvalidSkipIndex("Skip index invalid."))
                return

        if len(self.queue[ctx.guild.id].queue) > skip_index:
            if skip_index > 0:
                removed_songs = self.queue[ctx.guild.id].queue[:skip_index]

                self.queue[ctx.guild.id].queue = self.queue[ctx.guild.id].queue[skip_index:]
                if self.queue[ctx.guild.id].queue_loop:
                    self.queue[ctx.guild.id].queue += removed_songs

            player = self.queue[ctx.guild.id].queue[0]
            ctx.voice_client.stop()
            return player

        await self.call_event('on_music_error', ctx, SkipError("No song to skip to."))

    async def volume(self, ctx, volume: int = None):
        if not await self.__check_connection(ctx, True, check_queue=True):
            return

        if volume is None:
            return ctx.voice_client.source.volume * 100

        ctx.voice_client.source.volume = volume / 100
        self.queue[ctx.guild.id].volume = volume / 100
        return ctx.voice_client.source.volume * 100

    async def join(self, ctx):
        if ctx.voice_client and ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx,
                                  AlreadyConnected("Client is already connected to a voice channel"))
            return

        if not ctx.author.voice:
            await self.call_event('on_music_error', ctx, UserNotConnected("User is not connected to a voice channel"))
            return

        await ctx.author.voice.channel.connect()
        return True

    async def leave(self, ctx):
        if not await self.__check_connection(ctx):
            return

        await ctx.voice_client.disconnect()
        return True

    async def history(self, ctx):
        if not await self.__check_connection(ctx, check_queue=True):
            return

        return self.queue[ctx.guild.id].history

    async def now_playing(self, ctx):
        if not await self.__check_connection(ctx, check_playing=True, check_queue=True):
            return

        return self.queue[ctx.guild.id].now_playing

    async def queueloop(self, ctx):
        if not await self.__check_connection(ctx, check_playing=True, check_queue=True):
            return

        self.queue[ctx.guild.id].loop = Loops.QUEUE_LOOP if self.queue[ctx.guild.id].loop != Loops.QUEUE_LOOP else\
            Loops.NO_LOOP

        if self.queue[ctx.guild.id].loop == Loops.QUEUE_LOOP:
            self.queue[ctx.guild.id].add(self.queue[ctx.guild.id].now_playing)

        return self.queue[ctx.guild.id].loop == Loops.QUEUE_LOOP

    async def loop(self, ctx):
        if not await self.__check_connection(ctx, check_playing=True, check_queue=True):
            return

        self.queue[ctx.guild.id].loop = Loops.LOOP if self.queue[ctx.guild.id].loop != Loops.LOOP else Loops.NO_LOOP
        return self.queue[ctx.guild.id].loop == Loops.LOOP

    def get_queue(self, ctx):
        return self.queue[ctx.guild.id].queue



import discord
from discord.ext import commands
import discordSuperUtils

intents = discord.Intents.all()
bot = commands.Bot(command_prefix="-", intents=intents)

ImageManager = discordSuperUtils.ImageManager(bot, card_back=3, txt_colour=(255, 64, 99), custom_card_back=False)


@bot.event
async def on_ready():
    print(bot.user)


@bot.command()
async def gay(ctx, member: discord.Member = None):
    member = ctx.author if not member else member
    img = await ImageManager.add_gay(member.avatar_url)
    await ctx.send(file=img)


@bot.command()
async def merge(ctx, member1: discord.Member, member2: discord.Member):
    img = await ImageManager.merge_image(member1.avatar_url, member2.avatar_url, if_url=True)
    await ctx.send(file=img)


@bot.command()
async def testrank(ctx, member: discord.Member = None):
    member = ctx.author if not member else member
    img = await ImageManager.create_profile(member, 1, 100, 1230000, 2500000, 100000)
    await ctx.send(file=img)


bot.run("token")



import PIL
import PIL.ImageShow
import aiohttp  # koyashie fix unused / duplicate imports
import discord
import os
from PIL import Image, ImageFont, ImageDraw
from io import BytesIO


class ImageManager:
    def __init__(self, bot, txt_colour=None, card_back=1, custom_card_back: bool = False):
        self.bot = bot
        txt_colour = (80, 92, 112) if not txt_colour else txt_colour
        self.txt_colour = txt_colour  # tuple with rgb colour
        self.default_bg = self.fetch_card_back(card_back, custom_card_back)
        self.online = self.load_asset('online.png')
        self.offline = self.load_asset('offline.png')
        self.idle = self.load_asset('idle.png')
        self.dnd = self.load_asset('dnd.png')
        self.streaming = self.load_asset('streaming.png')
        self.font = self.load_asset('font.ttf')
        self.bk = self.load_asset('grey.png')

    @classmethod
    def load_asset(cls, name):
        return os.path.join(os.path.dirname(__file__), 'assets', name)

    @classmethod
    def fetch_card_back(cls, card_back, custom_card_back):
        if custom_card_back:
            return card_back

        if card_back in [1, 2, 3]:
            return cls.load_asset(f"{card_back}.png")

    @classmethod
    async def make_request(cls, url):
        async with aiohttp.ClientSession() as session:
            async with session.get(str(url)) as response:
                return await response.read()

    @classmethod
    async def convert_image(cls, url):
        info = await cls.make_request(url)
        return PIL.Image.open(BytesIO(info)).convert('RGBA')

    @staticmethod
    def create_card():
        """Solely for testing | Creates a blank image"""
        img = PIL.Image.new('RGB', (900, 238), color=(91, 95, 102))

    @classmethod
    def human_format(cls, num):
        original_num = num

        num = float('{:.3g}'.format(num))
        magnitude = 0
        matches = ['', 'K', 'M', 'B', 'T', 'Qua', 'Qui']
        while abs(num) >= 1000:
            if magnitude >= 5:
                break

            magnitude += 1
            num /= 1000.0

        try:
            return '{}{}'.format('{:f}'.format(num).rstrip('0').rstrip('.'), matches[magnitude])
        except IndexError:
            return original_num

    async def add_gay(self, avatar, discord_file: bool = True, if_url : bool = False):
        """Adds gay overlay to image url given"""
        gay_image = PIL.Image.open(self.load_asset('gay.jpg'))
        if if_url:
            avatar = await self.convert_image(avatar)

        width, height = avatar.size
        foreground = gay_image.convert('RGBA').resize((width, height), PIL.Image.ANTIALIAS)
        img = discord.File(await self.merge_image(foreground, avatar, 0.4))

        if discord_file:
            return discord.File(img, filename="Gay.png")
        return img

    async def merge_image(self, foreground, background, if_url: bool = False, blend_level: float = 0.6,
                          discord_file: bool = True):
        """Merges two images together"""
        if if_url:
            foreground = self.convert_image(foreground)
            background = self.convert_image(background)
        result_bytes = BytesIO()
        width, height = background.size

        foreground = foreground.resize((width, height), PIL.Image.ANTIALIAS)
        result = PIL.Image.blend(background, foreground, alpha=blend_level)

        if discord_file:
            result.save(result_bytes, format="PNG")
            result_bytes.seek(0)
            return discord.File(result_bytes, filename="mergedimage.png")

        return result

    async def create_profile(self, user: discord.Member, rank: int, level: int, xp: int, next_level_xp: int = None,
                             current_level_xp: int = None, discord_file=True):

        avatar = PIL.Image.open(BytesIO(await self.make_request(str(user.avatar_url))))
        avatar = avatar.convert('RGBA').resize((180, 180))
        card = Image.open(self.default_bg)
        card = card.resize((900, 238))
        font = ImageFont.truetype(self.font, 36)
        font_small = ImageFont.truetype(self.font, 20)
        result_bytes = BytesIO()

        status = Image.open(getattr(self, user.status.name))

        status = status.convert("RGBA").resize((55, 55))
        profile_pic_holder = Image.new("RGBA", card.size, (255, 255, 255, 0))

        mask = Image.new("RGBA", card.size, 0)
        mask_draw = ImageDraw.Draw(mask)
        mask_draw.ellipse(
            (29, 29, 209, 209), fill=(255, 25, 255, 255)
        )

        draw = ImageDraw.Draw(card)
        draw.text((245, 60), f"{user}", self.txt_colour, font=font)
        draw.text((620, 60), f"Rank #{rank}", self.txt_colour, font=font)
        draw.text((245, 145), f"Level {level}", self.txt_colour, font=font_small)
        draw.text((620, 145), f"{self.human_format(xp)} / {self.human_format(next_level_xp)} XP", self.txt_colour,
                  font=font_small)

        blank = Image.new("RGBA", card.size, (255, 255, 255, 0))
        blankdraw = ImageDraw.Draw(blank)
        blankdraw.rounded_rectangle((245, 185, 750, 205), fill=(255, 255, 255, 0), outline=self.txt_colour, radius=10)

        xpneed = next_level_xp - current_level_xp
        xphave = xp - current_level_xp
        length_of_bar = (((xphave / xpneed) * 100) * 4.9) + 248

        blankdraw.rounded_rectangle((248, 188, length_of_bar, 202), fill=self.txt_colour, radius=7)

        profile_pic_holder.paste(avatar, (29, 29, 209, 209))
        precard = Image.composite(profile_pic_holder, card, mask)
        precard = precard.convert('RGBA')
        precard = Image.alpha_composite(precard, blank)

        blank = Image.new("RGBA", card.size, (255, 255, 255, 0))
        blank.paste(status, (155, 155))
        finalcard = Image.alpha_composite(precard, blank)

        if discord_file:
            finalcard.save(result_bytes, format="PNG")
            result_bytes.seek(0)
            return discord.File(result_bytes, filename="rankcard.png")

        return finalcard

    async def greyscale(self, img, if_url: bool = False, discord_file: bool = False):
        if if_url:
            img = await self.convert_image(img)
        if discord_file:
            return discord.File(img.convert('LA'))
        return img.convert('LA')



import discordSuperUtils
from discord.ext import commands
import discord

bot = commands.Bot(command_prefix='-')
MusicManager = discordSuperUtils.MusicManager(bot)


@MusicManager.event()
async def on_music_error(ctx, error):
    raise error  # add your error handling here! Errors are listed in the documentation.


@MusicManager.event()
async def on_play(ctx, player):
    await ctx.send(f"Playing {player}")


@bot.event
async def on_ready():
    print('Music manager is ready.', bot.user)


@bot.command()
async def leave(ctx):
    if await MusicManager.leave(ctx):
        await ctx.send("Left Voice Channel")


@bot.command()
async def np(ctx):
    if player := await MusicManager.now_playing(ctx):
        await ctx.send(f"Currently playing: {player}")


@bot.command()
async def join(ctx):
    if await MusicManager.join(ctx):
        await ctx.send("Joined Voice Channel")


@bot.command()
async def play(ctx, *, query: str):
    player = await MusicManager.create_player(query)
    await MusicManager.queue_add(player=player, ctx=ctx)

    if not await MusicManager.play(ctx):
        await ctx.send("Added to queue")


@bot.command()
async def volume(ctx, volume: int):
    await MusicManager.volume(ctx, volume)


@bot.command()
async def loop(ctx):
    is_loop = await MusicManager.loop(ctx)
    await ctx.send(f"Looping toggled to {is_loop}")


@bot.command()
async def queueloop(ctx):
    is_loop = await MusicManager.queueloop(ctx)
    await ctx.send(f"Queue looping toggled to {is_loop}")


@bot.command()
async def history(ctx):
    embeds = discordSuperUtils.generate_embeds(await MusicManager.history(ctx),
                                               "Song History",
                                               "Shows all played songs",
                                               25,
                                               string_format="Title: {}")

    page_manager = discordSuperUtils.PageManager(ctx, embeds, public=True)
    await page_manager.run()


@bot.command()
async def skip(ctx, index: int = None):
    await MusicManager.skip(ctx, index)


@bot.command()
async def queue(ctx):
    embeds = discordSuperUtils.generate_embeds(MusicManager.get_queue(ctx),
                                               "Queue",
                                               f"Now Playing: {await MusicManager.now_playing(ctx)}",
                                               25,
                                               string_format="Title: {}")

    page_manager = discordSuperUtils.PageManager(ctx, embeds, public=True)
    await page_manager.run()


bot.run("token")



import asyncio


class EventManager:
    def __init__(self):
        self.events = {}

    async def call_event(self, name, *args, **kwargs):
        if name in self.events:
            for event in self.events[name]:
                await event(*args, **kwargs)

    def event(self, name=None):
        def inner(func):
            self.add_event(func, name)
            return func

        return inner

    def add_event(self, func, name=None):
        name = func.__name__ if not name else name

        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be async.')

        if name in self.events:
            self.events[name].append(func)
        else:
            self.events[name] = [func]

    def remove_event(self, func, name=None):
        name = func.__name__ if not name else name

        if name in self.events:
            self.events[name].remove(func)


import discordSuperUtils
import sqlite3
from discord.ext import commands


database = discordSuperUtils.DatabaseManager(sqlite3.connect("database"))
bot = commands.Bot(command_prefix='-')
EconomyManager = discordSuperUtils.EconomyManager(database, 'economy', bot)


@bot.event
async def on_ready():
    print('Economy manager is ready.', bot.user)


@bot.command()
async def create_account(ctx):
    await EconomyManager.create_account(ctx.author)  # wont create an account if there already is one.
    await ctx.send('Created account.')


@bot.command()
async def beg(ctx):
    account = await EconomyManager.get_account(ctx.author)
    account.change_currency(5)
    await ctx.send("You begged for cash and someone gave you 5 dollars!")


@bot.command()
async def leaderboard(ctx):
    guild_leaderboard = await EconomyManager.get_leaderboard(ctx.guild)
    formatted_leaderboard = [f"Member: {x.member}, Network: {x.net}" for x in guild_leaderboard]

    await discordSuperUtils.PageManager(ctx, discordSuperUtils.generate_embeds(
        formatted_leaderboard,
        title="Economy Leaderboard",
        fields=25,
        description=f"Leaderboard of {ctx.guild}"
    )).run()


bot.run("token")



from setuptools import setup

f = open("README.md", "r")
README = f.read()


setup(
  name = 'discordSuperUtils',
  packages = ['discordSuperUtils'],
  version = '0.1.2',
  license='MIT',
  description = 'Discord Bot Development made easy!',
  long_description=README,
  long_description_content_type="text/markdown",
  author = 'koyashie07 and adam7100',
  url = 'https://github.com/koyashie07/discord-super-utils',
  download_url = 'https://github.com/discordsuperutils/discord-super-utils/archive/refs/tags/0.1.1.tar.gz',
  keywords = ['discord', 'easy', 'discord.py', 'music', 'download', 'links', 'images', 'videos','audio', 'bot','paginator','economy','reaction','reaction roles'],
  install_requires=[
          'youtube-dl',
          'discord.py'
      ],
  classifiers=[  # Optional
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    'Development Status :: 4 - Beta',

    # Indicate who your project is intended for
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',

    # Pick your license as you wish
    'License :: OSI Approved :: MIT License',

    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)



from .Database import DatabaseManager
import time
import math
from .Base import EventManager
import asyncio


database_keys = ['guild', 'member', 'rank', 'xp', 'level_up']


class LevelingAccount:
    def __init__(self, database: DatabaseManager, table, guild: int, member: int, rank_multiplier=1.5):
        self.database = database
        self.table = table
        self.guild = guild
        self.member = member
        self.rank_multiplier = rank_multiplier

    def __str__(self):
        return f"<Account MEMBER={self.member}, GUILD={self.guild}>"

    def __repr__(self):
        return f"<Account GUILD={self.guild}, MEMBER={self.member}, XP={self.xp}, LEVEL={self.level}>"

    def __lt__(self, other):
        return self.net < other.net

    @property
    def __checks(self):
        return LevelingManager.generate_checks(self.guild, self.member)

    @property
    def xp(self):
        return self.database.select(['xp'], self.table, self.__checks)[0]

    @property
    def level(self):
        return self.database.select(['rank'], self.table, self.__checks)[0]

    @property
    def next_level(self):
        return self.database.select(['level_up'], self.table, self.__checks)[0]

    @property
    def percentage_next_level(self):
        initial_rank_xp = 0 if self.level_up == 50 else self.level_up / self.rank_multiplier
        return math.floor(abs(self.xp - initial_rank_xp) / (self.level_up - initial_rank_xp) * 100)

    @xp.setter
    def xp(self, value):
        self.database.update(['xp'], [value], self.table, self.__checks)

    @level.setter
    def level(self, value):
        self.database.update(['rank'], [value], self.table, self.__checks)

    @next_level.setter
    def next_level(self, value):
        self.database.update(['level_up'], [value], self.table, self.__checks)


class LevelingManager(EventManager):
    def __init__(self, database: DatabaseManager, table, bot, xp_on_message=5, rank_multiplier=1.5, xp_cooldown=60):
        super().__init__()
        self.database = database
        self.table = table
        self.bot = bot
        self.xp_on_message = xp_on_message
        self.rank_multiplier = rank_multiplier
        self.xp_cooldown = xp_cooldown

        self.cooldown_members = {}
        self.bot.add_listener(self.__handle_experience, "on_message")
        self.database.createtable(self.table, [{'name': key, 'type': 'INTEGER'} for key in database_keys], True)

    @staticmethod
    def generate_checks(guild: int, member: int):
        return [{'guild': guild}, {'member': member}]

    async def __handle_experience(self, message):
        if not message.guild or message.author.bot:
            return

        if message.guild.id not in self.cooldown_members:
            self.cooldown_members[message.guild.id] = {}

        self.create_account(message.author)
        member_account = self.get_account(message.author)
        member_timestamp = self.cooldown_members[message.guild.id].get(message.author.id, 0)

        if (time.time() - member_timestamp) >= self.xp_cooldown:
            member_account.xp += self.xp_on_message
            self.cooldown_members[message.guild.id][message.author.id] = time.time()

            leveled_up = False
            while member_account.xp >= member_account.next_level:
                member_account.next_level *= member_account.rank_multiplier
                member_account.level += 1
                leveled_up = True

            if leveled_up:
                loop = asyncio.get_event_loop()
                loop.create_task(self.call_event('on_level_up', message, member_account))

    def create_account(self, member):
        self.database.insertifnotexists(database_keys, [member.guild.id, member.id, 1, 0, 50],
                                        self.table, [{'guild': member.guild.id}, {'member': member.id}])

    def get_account(self, member):
        member_data = self.database.select([], self.table, self.generate_checks(member.guild.id, member.id), True)

        if member_data:
            return LevelingAccount(self.database, self.table, member.guild.id, member.id, self.rank_multiplier)

        return None

    def get_leaderboard(self, guild):
        guild_info = self.database.select([], self.table, [{'guild': guild.id}], True)
        members = [LevelingAccount(self.database, self.table, *member_info[:2], self.rank_multiplier)
                   for member_info in guild_info]

        members.sort()
        return members
